﻿using System;
using System.IO;
using System.Text;
using System.Xml;
using System.Xml.Xsl;
using HIPS.CommonSchemas.PatientIdentifier;
using HIPS.PcehrSchemas;
using System.Text.RegularExpressions;

namespace HIPS.Web.UI.Helpers
{
    /// <summary>
    /// Extension methods to assist with the PCEHR View feature.
    /// </summary>
    public static class PcehrViewExtensions
    {
        #region Methods

        public static string TransformToHtml(this byte[] doc, string stylesheetContent)
        {
            return _TransformToHtml(doc, stylesheetContent);
        }

        public static string TransformToHtml(this LocalClinicalDocumentResponse<PatientIdentifierBase> doc, string stylesheetContent)
        {
            return _TransformToHtml(doc.Document, stylesheetContent);
        }

        public static string TransformToHtml(this DocumentResponse doc, string stylesheetContent)
        {
            return _TransformToHtml(doc.Document, stylesheetContent);
        }

        private static string _TransformToHtml(byte[] document, string stylesheetContent)
        {
            // Load Xml Document
            XmlDocument xmlDoc = new XmlDocument();
            xmlDoc.PreserveWhitespace = true;
            using (MemoryStream ms = new MemoryStream(document))
            {
                xmlDoc.Load(ms);
            }

            // Load stylesheet
            var stylesheet = new XslCompiledTransform();
            XmlDocument styleDoc = new XmlDocument();
            styleDoc.PreserveWhitespace = true;
            styleDoc.LoadXml(stylesheetContent);

            // Stylesheet additions

            string cssAdditions = @"
                img { max-width:100% !important; max-height: 600px !important; } /* Stops wide images breaking outside document frame */
                .spaceToScrollToFootnotesDiv { display: none !important; } /* Unnecessary additional footer space */
            ";

            // Add CSS
            var cssHeadMediaAll = styleDoc.SelectSingleNode("//*[local-name()='template' and @name='addCSS']/*[local-name()='element' and @name='style'][1]");
            cssHeadMediaAll.AppendChild(styleDoc.CreateTextNode(cssAdditions));

            // Complete stylesheet load with additions
            stylesheet.Load(styleDoc);

            // Apply stylesheet
            StringBuilder sb = new StringBuilder();
            using (StringWriter writer = new StringWriter(sb))
            {
                stylesheet.Transform(xmlDoc, null, writer);
            }

            return sb.ToString();
        }

        public static string TransformToHtml(this ViewResponse view, string stylesheetContent, string documentViewUrlRoot, string patientId, DateTime fromDate, DateTime toDate, string groupBy)
        {
            // TODO: Unify some logic with DocumentResponse SaveAndTransform

            // Load Xml Document
            XmlDocument xmlDoc = new XmlDocument();
            xmlDoc.PreserveWhitespace = true;
            using (MemoryStream ms = new MemoryStream(view.Document))
            {
                xmlDoc.Load(ms);
            }

            // Load stylesheet
            var stylesheet = new XslCompiledTransform();
            stylesheet.Load(XmlReader.Create(new StringReader(stylesheetContent)));

            // Parameters
            XsltArgumentList arguments = new XsltArgumentList();
            arguments.AddParam("addJavaScript", string.Empty, "false"); // Disabled due to iframe rendering issues
            arguments.AddParam("collapsedGroups", string.Empty, "false"); // Enabled to support replacement javascript
            arguments.AddParam("groupBy", string.Empty, groupBy);

            // Apply stylesheet
            StringBuilder sb = new StringBuilder();
            using (StringWriter writer = new StringWriter(sb))
            {
                stylesheet.Transform(xmlDoc, arguments, writer);
            }

            // Rewrite document URLs
            string input = sb.ToString();
            string content = input;

            // Find all document urls
            Regex regx = new Regex(@"href=\""pcehr:(\d+\.)*(\d+)\/(\d+\.)*(\d+)", RegexOptions.IgnoreCase);

            MatchCollection matches = regx.Matches(input);
            foreach (Match match in matches)
            {
                // For each document url change the url to match the local path and add the selected date range.
                string url = string.Format(@"target=""_blank"" href=""{0}", documentViewUrlRoot);
                string replaceValue = string.Format(@"{0}?fromDate={1}&toDate={2}", match.Value.Replace("href=\"pcehr:", url), fromDate.ToShortDateString(), toDate.ToShortDateString());
                content = content.Replace(match.Value, replaceValue);
            }

            return content;
        }

        #endregion Methods
    }
}